@@ -28,11 +28,16 @@
 #include "config.h"
 #endif
 
+#define _GNU_SOURCE 1
 #include "libwacomint.h"
+#include "util-strings.h"
+#include <linux/input-event-codes.h>
 
+#include <assert.h>
 #include <glib.h>
 #include <dirent.h>
 #include <string.h>
+#include <stdbool.h>
 #include <stdlib.h>
 #include <stdio.h>
 
@@ -48,64 +53,82 @@ libwacom_class_string_to_enum(const char *class)
 	if (class == NULL || *class == '\0')
 		return WCLASS_UNKNOWN;
 
-	if (strcmp(class, "Intuos3") == 0)
+	if (streq(class, "Intuos3"))
 		return WCLASS_INTUOS3;
-	if (strcmp(class, "Intuos4") == 0)
+	if (streq(class, "Intuos4"))
 		return WCLASS_INTUOS4;
-	if (strcmp(class, "Intuos5") == 0)
+	if (streq(class, "Intuos5"))
 		return WCLASS_INTUOS5;
-	if (strcmp(class, "Cintiq") == 0)
+	if (streq(class, "Cintiq"))
 		return WCLASS_CINTIQ;
-	if (strcmp(class, "Bamboo") == 0)
+	if (streq(class, "Bamboo"))
 		return WCLASS_BAMBOO;
-	if (strcmp(class, "Graphire") == 0)
+	if (streq(class, "Graphire"))
 		return WCLASS_GRAPHIRE;
-	if (strcmp(class, "Intuos") == 0)
+	if (streq(class, "Intuos"))
 		return WCLASS_INTUOS;
-	if (strcmp(class, "Intuos2") == 0)
+	if (streq(class, "Intuos2"))
 		return WCLASS_INTUOS2;
-	if (strcmp(class, "ISDV4") == 0)
+	if (streq(class, "ISDV4"))
 		return WCLASS_ISDV4;
-	if (strcmp(class, "PenDisplay") == 0)
+	if (streq(class, "PenDisplay"))
 		return WCLASS_PEN_DISPLAYS;
-	if (strcmp(class, "Remote") == 0)
+	if (streq(class, "Remote"))
 		return WCLASS_REMOTE;
 
 	return WCLASS_UNKNOWN;
 }
 
-WacomStylusType
+static WacomStylusType
 type_from_str (const char *type)
 {
 	if (type == NULL)
 		return WSTYLUS_UNKNOWN;
-	if (strcmp (type, "General") == 0)
+	if (streq(type, "General"))
 		return WSTYLUS_GENERAL;
-	if (strcmp (type, "Inking") == 0)
+	if (streq(type, "Inking"))
 		return WSTYLUS_INKING;
-	if (strcmp (type, "Airbrush") == 0)
+	if (streq(type, "Airbrush"))
 		return WSTYLUS_AIRBRUSH;
-	if (strcmp (type, "Classic") == 0)
+	if (streq(type, "Classic"))
 		return WSTYLUS_CLASSIC;
-	if (strcmp (type, "Marker") == 0)
+	if (streq(type, "Marker"))
 		return WSTYLUS_MARKER;
-	if (strcmp (type, "Stroke") == 0)
+	if (streq(type, "Stroke"))
 		return WSTYLUS_STROKE;
-	if (strcmp (type, "Puck") == 0)
+	if (streq(type, "Puck"))
 		return WSTYLUS_PUCK;
+	if (streq(type, "3D"))
+		return WSTYLUS_3D;
+	if (streq(type, "Mobile"))
+		return WSTYLUS_MOBILE;
 	return WSTYLUS_UNKNOWN;
 }
 
+static WacomEraserType
+eraser_type_from_str (const char *type)
+{
+	if (type == NULL)
+		return WACOM_ERASER_NONE;
+	if (streq(type, "None"))
+		return WACOM_ERASER_NONE;
+	if (streq(type, "Invert"))
+		return WACOM_ERASER_INVERT;
+	if (streq(type, "Button"))
+		return WACOM_ERASER_BUTTON;
+	return WACOM_ERASER_UNKNOWN;
+}
+
 WacomBusType
 bus_from_str (const char *str)
 {
-	if (strcmp (str, "usb") == 0)
+	if (streq(str, "usb"))
 		return WBUSTYPE_USB;
-	if (strcmp (str, "serial") == 0)
+	if (streq(str, "serial"))
 		return WBUSTYPE_SERIAL;
-	if (strcmp (str, "bluetooth") == 0)
+	if (streq(str, "bluetooth"))
 		return WBUSTYPE_BLUETOOTH;
-	if (strcmp (str, "i2c") == 0)
+	if (streq(str, "i2c"))
 		return WBUSTYPE_I2C;
 	return WBUSTYPE_UNKNOWN;
 }
@@ -142,14 +165,13 @@ make_match_string (const char *name, WacomBusType bus, int vendor_id, int produc
 static gboolean
 match_from_string(const char *str, WacomBusType *bus, int *vendor_id, int *product_id, char **name)
 {
-	int rc = 1;
-	char busstr[64], namestr[64];
+	int rc = 1, len = 0;
+	char busstr[64];
 
-	memset(namestr, 0, sizeof(namestr));
-
-	rc = sscanf(str, "%63[^:]:%x:%x:%63c", busstr, vendor_id, product_id, namestr);
-	if (rc == 4) {
-		*name = g_strdup(namestr);
+	rc = sscanf(str, "%63[^:]:%x:%x:%n", busstr, vendor_id, product_id, &len);
+	if (len > 0) {
+		/* Grumble grumble scanf handling of %n */
+		*name = g_strdup(str+len);
 	} else if (rc == 3) {
 		*name = NULL;
 	} else {
@@ -171,7 +193,7 @@ libwacom_matchstr_to_match(WacomDevice *device, const char *matchstr)
 	if (matchstr == NULL)
 		return FALSE;
 
-	if (g_strcmp0 (matchstr, GENERIC_DEVICE_MATCH) == 0) {
+	if (streq(matchstr, GENERIC_DEVICE_MATCH)) {
 		name = NULL;
 		bus = WBUSTYPE_UNKNOWN;
 		vendor_id = 0;
@@ -182,8 +204,8 @@ libwacom_matchstr_to_match(WacomDevice *device, const char *matchstr)
 	}
 
 	match = libwacom_match_new(name, bus, vendor_id, product_id);
-	libwacom_update_match(device, match);
-	libwacom_match_destroy(match);
+	libwacom_add_match(device, match);
+	libwacom_match_unref(match);
 
 	free(name);
 	return TRUE;
@@ -230,39 +252,50 @@ libwacom_parse_stylus_keyfile(WacomDeviceDatabase *db, const char *path)
 		int id;
 		char **string_list;
 
-		id = strtol (groups[i], NULL, 16);
-		if (id == 0) {
+		if (!safe_atoi_base (groups[i], &id, 16)) {
 			g_warning ("Failed to parse stylus ID '%s'", groups[i]);
 			continue;
 		}
 
 		stylus = g_new0 (WacomStylus, 1);
+		stylus->refcnt = 1;
 		stylus->id = id;
 		stylus->name = g_key_file_get_string(keyfile, groups[i], "Name", NULL);
+		stylus->group = g_key_file_get_string(keyfile, groups[i], "Group", NULL);
 
-		stylus->is_eraser = g_key_file_get_boolean(keyfile, groups[i], "IsEraser", &error);
+		type = g_key_file_get_string(keyfile, groups[i], "EraserType", NULL);
+		stylus->eraser_type = eraser_type_from_str (type);
+		g_free (type);
+
+		string_list = g_key_file_get_string_list (keyfile, groups[i], "PairedStylusIds", NULL, NULL);
+		if (string_list) {
+			GArray *array;
+			guint j;
+
+			array = g_array_new (FALSE, FALSE, sizeof(int));
+			for (j = 0; string_list[j]; j++) {
+				int val;
+
+				if (safe_atoi_base (string_list[j], &val, 16)) {
+					g_array_append_val (array, val);
+					stylus->num_ids++;
+				} else {
+					g_warning ("Stylus %s (%s) Ignoring invalid PairedId value\n", stylus->name, groups[i]);
+				}
+			}
+
+			g_strfreev (string_list);
+			stylus->paired_ids = (int *) g_array_free (array, FALSE);
+		}
+
+		stylus->has_lens = g_key_file_get_boolean(keyfile, groups[i], "HasLens", &error);
+		if (error && error->code == G_KEY_FILE_ERROR_INVALID_VALUE)
+			g_warning ("Stylus %s (%s) %s\n", stylus->name, groups[i], error->message);
+		g_clear_error (&error);
+		stylus->has_wheel = g_key_file_get_boolean(keyfile, groups[i], "HasWheel", &error);
 		if (error && error->code == G_KEY_FILE_ERROR_INVALID_VALUE)
 			g_warning ("Stylus %s (%s) %s\n", stylus->name, groups[i], error->message);
 		g_clear_error (&error);
-
-		if (stylus->is_eraser == FALSE) {
-			stylus->has_eraser = g_key_file_get_boolean(keyfile, groups[i], "HasEraser", &error);
-			if (error && error->code == G_KEY_FILE_ERROR_INVALID_VALUE)
-				g_warning ("Stylus %s (%s) %s\n", stylus->name, groups[i], error->message);
-			g_clear_error (&error);
-			stylus->has_lens = g_key_file_get_boolean(keyfile, groups[i], "HasLens", &error);
-			if (error && error->code == G_KEY_FILE_ERROR_INVALID_VALUE)
-				g_warning ("Stylus %s (%s) %s\n", stylus->name, groups[i], error->message);
-			g_clear_error (&error);
-			stylus->has_wheel = g_key_file_get_boolean(keyfile, groups[i], "HasWheel", &error);
-			if (error && error->code == G_KEY_FILE_ERROR_INVALID_VALUE)
-				g_warning ("Stylus %s (%s) %s\n", stylus->name, groups[i], error->message);
-			g_clear_error (&error);
-		} else {
-			stylus->has_eraser = FALSE;
-			stylus->has_lens = FALSE;
-			stylus->has_wheel = FALSE;
-		}
 
 		stylus->num_buttons = g_key_file_get_integer(keyfile, groups[i], "Buttons", &error);
 		if (stylus->num_buttons == 0 && error != NULL) {
@@ -273,27 +306,27 @@ libwacom_parse_stylus_keyfile(WacomDeviceDatabase *db, const char *path)
 		string_list = g_key_file_get_string_list (keyfile, groups[i], "Axes", NULL, NULL);
 		if (string_list) {
 			WacomAxisTypeFlags axes = WACOM_AXIS_TYPE_NONE;
-			guint i;
+			guint j;
 
-			for (i = 0; string_list[i]; i++) {
+			for (j = 0; string_list[j]; j++) {
 				WacomAxisTypeFlags flag = WACOM_AXIS_TYPE_NONE;
-				if (strcmp (string_list[i], "Tilt") == 0) {
+				if (streq(string_list[j], "Tilt")) {
 					flag = WACOM_AXIS_TYPE_TILT;
-				} else if (strcmp (string_list[i], "RotationZ") == 0) {
+				} else if (streq(string_list[j], "RotationZ")) {
 					flag = WACOM_AXIS_TYPE_ROTATION_Z;
-				} else if (strcmp (string_list[i], "Distance") == 0) {
+				} else if (streq(string_list[j], "Distance")) {
 					flag = WACOM_AXIS_TYPE_DISTANCE;
-				} else if (strcmp (string_list[i], "Pressure") == 0) {
+				} else if (streq(string_list[j], "Pressure")) {
 					flag = WACOM_AXIS_TYPE_PRESSURE;
-				} else if (strcmp (string_list[i], "Slider") == 0) {
+				} else if (streq(string_list[j], "Slider")) {
 					flag = WACOM_AXIS_TYPE_SLIDER;
 				} else {
 					g_warning ("Invalid axis %s for stylus ID %s\n",
-						   string_list[i], groups[i]);
+						   string_list[j], groups[i]);
 				}
 				if (axes & flag)
 					g_warning ("Duplicate axis %s for stylus ID %s\n",
-						   string_list[i], groups[i]);
+						   string_list[j], groups[i]);
 				axes |= flag;
 			}
 
@@ -306,7 +339,7 @@ libwacom_parse_stylus_keyfile(WacomDeviceDatabase *db, const char *path)
 		g_free (type);
 
 		if (g_hash_table_lookup (db->stylus_ht, GINT_TO_POINTER (id)) != NULL)
-			g_warning ("Duplicate definition for stylus ID '0x%x'", id);
+			g_warning ("Duplicate definition for stylus ID '%#x'", id);
 
 		g_hash_table_insert (db->stylus_ht, GINT_TO_POINTER (id), stylus);
 	}
@@ -315,6 +348,35 @@ libwacom_parse_stylus_keyfile(WacomDeviceDatabase *db, const char *path)
 		g_key_file_free (keyfile);
 }
 
+static void
+libwacom_setup_paired_attributes(WacomDeviceDatabase *db)
+{
+	GHashTableIter iter;
+	gpointer key, value;
+
+	g_hash_table_iter_init(&iter, db->stylus_ht);
+	while (g_hash_table_iter_next (&iter, &key, &value)) {
+		WacomStylus *stylus = value;
+		const int* ids;
+		int count;
+		int i;
+
+		ids = libwacom_stylus_get_paired_ids(stylus, &count);
+		for (i = 0; i < count; i++) {
+			const WacomStylus *pair;
+
+			pair = libwacom_stylus_get_for_id(db, ids[i]);
+			if (pair == NULL) {
+				g_warning("Paired stylus '0x%x' not found, ignoring.", ids[i]);
+				continue;
+			}
+			if (libwacom_stylus_is_eraser(pair)) {
+				stylus->has_eraser = true;
+			}
+		}
+	}
+}
+
 struct {
 	const char       *key;
 	WacomButtonFlags  flag;
@@ -377,6 +439,95 @@ libwacom_parse_buttons_key(WacomDevice      *device,
 	g_strfreev (vals);
 }
 
+static inline bool
+set_button_codes_from_string(WacomDevice *device, char **strvals)
+{
+	gint i;
+
+	assert(strvals);
+
+	for (i = 0; i < device->num_buttons; i++) {
+		gint val;
+
+		if (!strvals[i]) {
+			g_error ("%s: Missing EvdevCode for button %d, ignoring all codes\n",
+				 device->name, i);
+			return false;
+		}
+
+		if (!safe_atoi_base (strvals[i], &val, 16) || val < BTN_MISC || val >= BTN_DIGI) {
+			g_warning ("%s: Invalid EvdevCode %s for button %d, ignoring all codes\n",
+				   device->name, strvals[i], i);
+			return false;
+		}
+		device->button_codes[i] = val;
+	}
+
+	return true;
+}
+
+static inline void
+set_button_codes_from_heuristics(WacomDevice *device)
+{
+	gint i;
+	for (i = 0; i < device->num_buttons; i++) {
+		if (device->cls == WCLASS_BAMBOO ||
+		    device->cls == WCLASS_GRAPHIRE) {
+			switch (i) {
+			case 0:
+				device->button_codes[i] = BTN_LEFT;
+				break;
+			case 1:
+				device->button_codes[i] = BTN_RIGHT;
+				break;
+			case 2:
+				device->button_codes[i] = BTN_FORWARD;
+				break;
+			case 3:
+				device->button_codes[i] = BTN_BACK;
+				break;
+			default:
+				device->button_codes[i] = 0;
+				break;
+			}
+		} else {
+			/* Assume traditional ExpressKey ordering */
+			switch (i) {
+			case 0 ... 9:
+				device->button_codes[i] = BTN_0 + i;
+				break;
+			case 10 ... 15:
+				device->button_codes[i] = BTN_A + (i-10);
+				break;
+			case 16:
+			case 17:
+				device->button_codes[i] = BTN_BASE + (i-16);
+				break;
+			default:
+				device->button_codes[i] = 0;
+				break;
+			}
+		}
+
+		if (device->button_codes[i] == 0)
+			g_warning ("Unable to determine evdev code for button %d (%s)", i, device->name);
+	}
+}
+
+static void
+libwacom_parse_button_codes(WacomDevice *device,
+			    GKeyFile    *keyfile)
+{
+	char **vals;
+
+	vals = g_key_file_get_string_list(keyfile, BUTTONS_GROUP, "EvdevCodes", NULL, NULL);
+	if (!vals ||
+	    !set_button_codes_from_string(device, vals))
+		set_button_codes_from_heuristics(device);
+
+	g_strfreev (vals);
+}
+
 static int
 libwacom_parse_num_modes (WacomDevice      *device,
 			  GKeyFile         *keyfile,
@@ -384,7 +535,7 @@ libwacom_parse_num_modes (WacomDevice      *device,
 			  WacomButtonFlags  flag)
 {
 	int num;
-	guint i;
+	int i;
 
 	num = g_key_file_get_integer (keyfile, BUTTONS_GROUP, key, NULL);
 	if (num > 0)
@@ -405,13 +556,65 @@ libwacom_parse_buttons(WacomDevice *device,
 	for (i = 0; i < G_N_ELEMENTS (options); i++)
 		libwacom_parse_buttons_key(device, keyfile, options[i].key, options[i].flag);
 
+	libwacom_parse_button_codes(device, keyfile);
+
 	device->ring_num_modes = libwacom_parse_num_modes(device, keyfile, "RingNumModes", WACOM_BUTTON_RING_MODESWITCH);
 	device->ring2_num_modes = libwacom_parse_num_modes(device, keyfile, "Ring2NumModes", WACOM_BUTTON_RING2_MODESWITCH);
 	device->strips_num_modes = libwacom_parse_num_modes(device, keyfile, "StripsNumModes", WACOM_BUTTON_TOUCHSTRIP_MODESWITCH);
 }
 
+static int
+styli_id_sort(gconstpointer pa, gconstpointer pb)
+{
+	const int *a = pa, *b = pb;
+	return *a > *b ? 1 : *a == *b ? 0 : -1;
+}
+
+static void
+libwacom_parse_styli_list(WacomDeviceDatabase *db, WacomDevice *device,
+			  char **ids)
+{
+	GArray *array;
+	guint i;
+
+	array = g_array_new (FALSE, FALSE, sizeof(int));
+	device->num_styli = 0;
+	for (i = 0; ids[i]; i++) {
+		const char *id = ids[i];
+
+		if (strneq(id, "0x", 2)) {
+			int int_value;
+			if (safe_atoi_base (ids[i], &int_value, 16)) {
+				g_array_append_val (array, int_value);
+				device->num_styli++;
+			}
+		} else if (strneq(id, "@", 1)) {
+			const char *group = &id[1];
+			GHashTableIter iter;
+			gpointer key, value;
+
+			g_hash_table_iter_init(&iter, db->stylus_ht);
+			while (g_hash_table_iter_next (&iter, &key, &value)) {
+				WacomStylus *stylus = value;
+				if (stylus->group && streq(group, stylus->group)) {
+					g_array_append_val (array, stylus->id);
+					device->num_styli++;
+				}
+			}
+		} else {
+			g_warning ("Invalid prefix for '%s'!", id);
+		}
+	}
+	/* Using groups means we don't get the styli in ascending order.
+	   Sort it so the output is predictable */
+	g_array_sort(array, styli_id_sort);
+	device->supported_styli = (int *) g_array_free (array, FALSE);
+}
+
 static WacomDevice*
-libwacom_parse_tablet_keyfile(const char *datadir, const char *filename)
+libwacom_parse_tablet_keyfile(WacomDeviceDatabase *db,
+			      const char *datadir,
+			      const char *filename)
 {
 	WacomDevice *device = NULL;
 	GKeyFile *keyfile;
@@ -422,6 +625,7 @@ libwacom_parse_tablet_keyfile(const char *datadir, const char *filename)
 	char *class;
 	char *paired;
 	char **string_list;
+	bool success = FALSE;
 
 	keyfile = g_key_file_new();
 
@@ -434,6 +638,7 @@ libwacom_parse_tablet_keyfile(const char *datadir, const char *filename)
 	}
 
 	device = g_new0 (WacomDevice, 1);
+	device->refcnt = 1;
 
 	string_list = g_key_file_get_string_list(keyfile, DEVICE_GROUP, "DeviceMatch", NULL, NULL);
 	if (!string_list) {
@@ -451,10 +656,8 @@ libwacom_parse_tablet_keyfile(const char *datadir, const char *filename)
 		}
 		if (nmatches == 0) {
 			DBG("failed to match '%s' for product/vendor IDs in '%s'\n", string_list[i], path);
-				g_strfreev (string_list);
-				g_free (device);
-				device = NULL;
-				goto out;
+			g_strfreev (string_list);
+			goto out;
 		}
 		if (nmatches > 1) {
 			/* set default to first entry */
@@ -470,6 +673,13 @@ libwacom_parse_tablet_keyfile(const char *datadir, const char *filename)
 	}
 
 	device->name = g_key_file_get_string(keyfile, DEVICE_GROUP, "Name", NULL);
+	device->model_name = g_key_file_get_string(keyfile, DEVICE_GROUP, "ModelName", NULL);
+	/* ModelName= would give us the empty string, let's make it NULL
+	 * instead */
+	if (device->model_name && strlen(device->model_name) == 0) {
+		free(device->model_name);
+		device->model_name = NULL;
+	}
 	device->width = g_key_file_get_integer(keyfile, DEVICE_GROUP, "Width", NULL);
 	device->height = g_key_file_get_integer(keyfile, DEVICE_GROUP, "Height", NULL);
 
@@ -483,7 +693,7 @@ libwacom_parse_tablet_keyfile(const char *datadir, const char *filename)
 		for (i = 0; string_list[i]; i++) {
 			found = FALSE;
 			for (n = 0; n < G_N_ELEMENTS (integration_flags); n++) {
-				if (!strcmp(string_list[i], integration_flags[n].key)) {
+				if (streq(string_list[i], integration_flags[n].key)) {
 					device->integration_flags |= integration_flags[n].value;
 					found = TRUE;
 					break;
@@ -508,24 +718,12 @@ libwacom_parse_tablet_keyfile(const char *datadir, const char *filename)
 
 	string_list = g_key_file_get_string_list(keyfile, DEVICE_GROUP, "Styli", NULL, NULL);
 	if (string_list) {
-		GArray *array;
-		guint i;
-
-		array = g_array_new (FALSE, FALSE, sizeof(int));
-		device->num_styli = 0;
-		for (i = 0; string_list[i]; i++) {
-			glong long_value = strtol (string_list[i], NULL, 0);
-			int int_value = long_value;
-
-			g_array_append_val (array, int_value);
-			device->num_styli++;
-		}
+		libwacom_parse_styli_list(db, device, string_list);
 		g_strfreev (string_list);
-		device->supported_styli = (int *) g_array_free (array, FALSE);
 	} else {
 		device->supported_styli = g_new (int, 2);
-		device->supported_styli[0] = WACOM_STYLUS_FALLBACK_ID;
-		device->supported_styli[1] = WACOM_ERASER_FALLBACK_ID;
+		device->supported_styli[0] = WACOM_ERASER_FALLBACK_ID;
+		device->supported_styli[1] = WACOM_STYLUS_FALLBACK_ID;
 		device->num_styli = 2;
 	}
 
@@ -570,6 +768,7 @@ libwacom_parse_tablet_keyfile(const char *datadir, const char *filename)
 	}
 	if (device->num_buttons > 0) {
 		device->buttons = g_new0 (WacomButtonFlags, device->num_buttons);
+		device->button_codes = g_new0 (gint, device->num_buttons);
 		libwacom_parse_buttons(device, keyfile);
 	}
 
@@ -582,7 +781,7 @@ libwacom_parse_tablet_keyfile(const char *datadir, const char *filename)
 		device->num_leds = 0;
 		for (i = 0; string_list[i]; i++) {
 			for (n = 0; n < G_N_ELEMENTS (supported_leds); n++) {
-				if (!strcmp(string_list[i], supported_leds[n].key)) {
+				if (streq(string_list[i], supported_leds[n].key)) {
 					g_array_append_val (array, supported_leds[n].value);
 					device->num_leds++;
 					break;
@@ -593,6 +792,8 @@ libwacom_parse_tablet_keyfile(const char *datadir, const char *filename)
 		device->status_leds = (WacomStatusLEDs *) g_array_free (array, FALSE);
 	}
 
+	success = TRUE;
+
 out:
 	if (path)
 		g_free(path);
@@ -600,138 +801,180 @@ out:
 		g_key_file_free(keyfile);
 	if (error)
 		g_error_free(error);
+	if (!success)
+		device = libwacom_unref(device);
 
 	return device;
 }
 
-static int
-scandir_tablet_filter(const struct dirent *entry)
+static bool
+has_suffix(const char *name, const char *suffix)
 {
-	const char *name = entry->d_name;
-	int len, suffix_len;
+	size_t len = strlen(name);
+	size_t suffix_len = strlen(suffix);
 
 	if (!name || name[0] == '.')
 		return 0;
 
-	len = strlen(name);
-	suffix_len = strlen(TABLET_SUFFIX);
 	if (len <= suffix_len)
-		return 0;
+		return false;
 
-	return !strcmp(&name[len - suffix_len], TABLET_SUFFIX);
+	return streq(&name[len - suffix_len], suffix);
 }
 
 static int
-scandir_stylus_filter(const struct dirent *entry)
+is_tablet_file(const struct dirent *entry)
 {
-	const char *name = entry->d_name;
-	int len, suffix_len;
-
-	if (!name || name[0] == '.')
-		return 0;
-
-	len = strlen(name);
-	suffix_len = strlen(STYLUS_SUFFIX);
-	if (len <= suffix_len)
-		return 0;
-
-	return !strcmp(&name[len - suffix_len], STYLUS_SUFFIX);
+	return has_suffix(entry->d_name, TABLET_SUFFIX);
 }
 
+static int
+is_stylus_file(const struct dirent *entry)
+{
+	return has_suffix(entry->d_name, STYLUS_SUFFIX);
+}
 
-WacomDeviceDatabase *
+static bool
+load_tablet_files(WacomDeviceDatabase *db, const char *datadir)
+{
+	DIR *dir;
+	struct dirent *file;
+	bool success = false;
+
+	dir = opendir(datadir);
+	if (!dir)
+		return errno == ENOENT; /* non-existing directory is ok */
+
+	while ((file = readdir(dir))) {
+		WacomDevice *d;
+		const WacomMatch **matches, **match;
+
+		if (!is_tablet_file(file))
+			continue;
+
+		d = libwacom_parse_tablet_keyfile(db, datadir, file->d_name);
+		if (!d)
+			continue;
+
+		matches = libwacom_get_matches(d);
+		if (!matches || !*matches) {
+			g_critical("Device '%s' has no matches defined\n",
+				   libwacom_get_name(d));
+			goto out;
+		}
+
+		for (match = matches; *match; match++) {
+			const char *matchstr;
+			matchstr = libwacom_match_get_match_string(*match);
+			/* no duplicate matches allowed */
+			if (g_hash_table_lookup(db->device_ht, matchstr) != NULL) {
+				g_critical("Duplicate match of '%s' on device '%s'.",
+					   matchstr, libwacom_get_name(d));
+				goto out;
+			}
+			g_hash_table_insert (db->device_ht, g_strdup (matchstr), d);
+			libwacom_ref(d);
+		}
+		libwacom_unref(d);
+	}
+
+	success = true;
+
+out:
+	closedir(dir);
+	return success;
+}
+
+static void
+stylus_destroy(void *data)
+{
+	libwacom_stylus_unref((WacomStylus*)data);
+}
+
+static bool
+load_stylus_files(WacomDeviceDatabase *db, const char *datadir)
+{
+	DIR *dir;
+	struct dirent *file;
+
+	dir = opendir(datadir);
+	if (!dir)
+		return errno == ENOENT; /* non-existing directory is ok */
+
+	while ((file = readdir(dir))) {
+		char *path;
+
+		if (!is_stylus_file(file))
+			continue;
+
+		path = g_build_filename (datadir, file->d_name, NULL);
+		libwacom_parse_stylus_keyfile(db, path);
+		g_free(path);
+	}
+
+	closedir(dir);
+
+	return true;
+}
+
+static bool
+load_from_datadir(WacomDeviceDatabase *db, const char *datadir)
+{
+	return load_stylus_files(db, datadir) && load_tablet_files(db, datadir);
+}
+
+static WacomDeviceDatabase *
+database_new_for_paths (size_t npaths, const char **datadirs)
+{
+	WacomDeviceDatabase *db;
+
+	db = g_new0 (WacomDeviceDatabase, 1);
+	db->device_ht = g_hash_table_new_full (g_str_hash,
+					       g_str_equal,
+					       g_free,
+					       (GDestroyNotify) libwacom_destroy);
+	db->stylus_ht = g_hash_table_new_full (g_direct_hash,
+					       g_direct_equal,
+					       NULL,
+					       (GDestroyNotify) stylus_destroy);
+
+	for (const char **datadir = datadirs; npaths--; datadir++) {
+		if (!load_from_datadir(db, *datadir))
+			goto error;
+	}
+
+	/* If we couldn't load _anything_ then something's wrong */
+	if (g_hash_table_size (db->stylus_ht) == 0 ||
+	    g_hash_table_size (db->device_ht) == 0)
+		goto error;
+
+	libwacom_setup_paired_attributes(db);
+
+	return db;
+
+error:
+	libwacom_database_destroy(db);
+	return NULL;
+}
+
+LIBWACOM_EXPORT WacomDeviceDatabase *
 libwacom_database_new_for_path (const char *datadir)
 {
-    int n, nfiles;
-    struct dirent **files;
-    WacomDeviceDatabase *db;
-    char *path;
-
-    db = g_new0 (WacomDeviceDatabase, 1);
-
-    /* Load tablets */
-    db->device_ht = g_hash_table_new_full (g_str_hash,
-					   g_str_equal,
-					   g_free,
-					   (GDestroyNotify) libwacom_destroy);
-
-    n = scandir(datadir, &files, scandir_tablet_filter, alphasort);
-    if (n <= 0) {
-	    libwacom_database_destroy(db);
-	    return NULL;
-    }
-
-    nfiles = n;
-    while(n--) {
-	    WacomDevice *d;
-	    const WacomMatch **matches, **match;
-
-	    d = libwacom_parse_tablet_keyfile(datadir, files[n]->d_name);
-
-	    if (!d)
-		    continue;
-
-	    matches = libwacom_get_matches(d);
-	    for (match = matches; *match; match++) {
-		    const char *matchstr;
-		    matchstr = libwacom_match_get_match_string(*match);
-		    /* no duplicate matches allowed */
-		    if (g_hash_table_lookup(db->device_ht, matchstr) != NULL) {
-			    g_critical("Duplicate match of '%s' on device '%s'.",
-					matchstr, libwacom_get_name(d));
-			    libwacom_database_destroy(db);
-			    db = NULL;
-			    goto end;
-		    }
-		    g_hash_table_insert (db->device_ht, g_strdup (matchstr), d);
-		    d->refcnt++;
-	    }
-    }
-
-    while(nfiles--)
-	    free(files[nfiles]);
-    free(files);
-
-    /* Load styli */
-    n = scandir(datadir, &files, scandir_stylus_filter, alphasort);
-    if (n <= 0) {
-	    libwacom_database_destroy(db);
-	    return NULL;
-    }
-
-    db->stylus_ht = g_hash_table_new_full (g_direct_hash,
-					   g_direct_equal,
-					   NULL,
-					   (GDestroyNotify) libwacom_stylus_destroy);
-    nfiles = n;
-    while(n--) {
-	    path = g_build_filename (datadir, files[n]->d_name, NULL);
-	    libwacom_parse_stylus_keyfile(db, path);
-	    g_free(path);
-    }
-
-    /* If we couldn't load _anything_ then something's wrong */
-    if (g_hash_table_size (db->device_ht) == 0 &&
-	g_hash_table_size (db->stylus_ht) == 0) {
-	    libwacom_database_destroy(db);
-	    db = NULL;
-    }
-
-end:
-    while(nfiles--)
-	    free(files[nfiles]);
-    free(files);
-
-    return db;
+	return database_new_for_paths(1, &datadir);
 }
 
-WacomDeviceDatabase *
+LIBWACOM_EXPORT WacomDeviceDatabase *
 libwacom_database_new (void)
 {
-	return libwacom_database_new_for_path (DATADIR);
+	const char *datadir[] = {
+		DATADIR,
+		ETCDIR,
+	};
+
+	return database_new_for_paths (2, datadir);
 }
 
-void
+LIBWACOM_EXPORT void
 libwacom_database_destroy(WacomDeviceDatabase *db)
 {
 	if (db->device_ht)
@@ -741,7 +984,20 @@ libwacom_database_destroy(WacomDeviceDatabase *db)
 	g_free (db);
 }
 
-WacomDevice**
+static gint
+device_compare(gconstpointer pa, gconstpointer pb)
+{
+	const WacomDevice *a = pa,
+		          *b = pb;
+	int cmp;
+
+	cmp = libwacom_get_vendor_id(a) - libwacom_get_vendor_id(b);
+	if (cmp == 0)
+		cmp = libwacom_get_product_id(a) - libwacom_get_product_id(b);
+	return cmp;
+}
+
+LIBWACOM_EXPORT WacomDevice**
 libwacom_list_devices_from_database(const WacomDeviceDatabase *db, WacomError *error)
 {
 	GList *cur, *devices;
@@ -756,9 +1012,12 @@ libwacom_list_devices_from_database(const WacomDeviceDatabase *db, WacomError *e
 	list = calloc (g_list_length (devices) + 1, sizeof (WacomDevice *));
 	if (!list) {
 		libwacom_error_set(error, WERROR_BAD_ALLOC, "Memory allocation failed");
+		g_list_free (devices);
 		return NULL;
 	}
 
+	devices = g_list_sort (devices, device_compare);
+
 	for (p = list, cur = devices; cur; cur = g_list_next (cur))
 		*p++ = (WacomDevice *) cur->data;
 	g_list_free (devices);
